/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "dynamic_array.h"
#include "test_utils.h"

#define DARRAY_NAME str
#define DARRAY_DATA const char*
#include "dynamic_array.h"

const char* strs[] = {
  "Rcvfbqr", "1,", "XARR-QRRC", "VA", "GUR", "QRNQ:\n",
  "---------------------------------", "BAPR", "LBH", "ORNG", "GUR", "OVT",
  "ONQNFFRF", "NAQ", "PYRNA", "BHG", "GUR", "ZBBA", "ONFR", "LBH'ER",
  "FHCCBFRQ", "GB\n", "JVA", "NERA'G", "LBH?", "NERA'G", "LBH?", "JURER'F",
  "LBHE", "SNG", "ERJNEQ", "NAQ", "GVPXRG", "UBZR?", "JUNG\n", "GUR", "URYY",
  "VF", "GUVF?", "VG'F", "ABG", "FHCCBFRQ", "GB", "RAQ", "GUVF", "JNL!", "VG",
  "FGVAXF", "YVXR", "EBGGRA", "ZRNG,", "OHG", "YBBXF", "YVXR", "GUR", "YBFG",
  "QRVZBF", "ONFR.", "YBBXF", "YVXR\n", "LBH'ER", "FGHPX", "BA", "GUR",
  "FUBERF", "BS", "URYY.", "GUR", "BAYL", "JNL", "BHG", "VF", "GUEBHTU.", "GB",
  "PBAGVAHR", "GUR", "QBBZ", "RKCREVRAPR,", "CYNL", "GUR", "FUBERF", "BS",
  "URYY", "NAQ", "VGF", "NZNMVAT\n", "FRDHRY,", "VASREAB!",

  "Rcvfbqr 2, GUR FUBERF BS URYY:\n\
  ------------------------------\n\
  \n\
  LBH'IR QBAR VG! GUR UVQRBHF PLORE- QRZBA YBEQ GUNG EHYRQ GUR YBFG QRVZBF ZBBA\n\
  ONFR UNF ORRA FYNVA NAQ LBH NER GEVHZCUNAG! OHG ... JURER NER LBH? LBH\n\
  PYNZORE GB GUR RQTR BS GUR ZBBA NAQ YBBX QBJA GB FRR GUR NJSHY GEHGU.\n\
  \n",

  "QRVZBF SYBNGF NOBIR URYY VGFRYS!  LBH'IR ARIRE URNEQ BS NALBAR RFPNCVAT SEBZ\n\
  URYY, OHG LBH'YY ZNXR GUR ONFGNEQF FBEEL GURL RIRE URNEQ BS LBH! DHVPXYL, LBH\n\
  ENCCRY QBJA GB GUR FHESNPR BS URYY.\n\
  \n\
  ABJ, VG'F BA GB GUR SVANY PUNCGRE BS QBBZ! -- VASREAB."
};

#if 0
static void
test_primitive_type(void)
{
  struct mem_allocator allocator_proxy;
  struct darray_str darray;
  const size_t nstrs = sizeof(strs)/sizeof(const char*);
  size_t i = 0;

  mem_init_proxy_allocator(&allocator_proxy, &mem_default_allocator);

  darray_str_init(NULL, &darray);
  CHECK(darray_str_size_get(&darray), 0);
  darray_str_clear(&darray);
  CHECK(darray_str_size_get(&darray), 0);

  FOR_EACH(i, 0, 4) {
    darray_str_push_back(&darray, strs + i);
  }
  CHECK(darray_str_size_get(&darray), 4);
  darray_str_clear(&darray);
  CHECK(darray_str_size_get(&darray), 0);
  darray_str_release(&darray);

  darray_str_init(&allocator_proxy, &darray);
  FOR_EACH(i, 0, nstrs) {
    darray_str_push_back(&darray, strs + i);
  }
  FOR_EACH(i, 0, nstrs) {
    CHECK(strcmp(darray_str_cdata_get(&darray)[i], strs[i]), 0);
  }
  darray_str_clear(&darray);
  CHECK(darray_str_size_get(&darray), 0);
  darray_str_release(&darray);

  darray_str_init(&allocator_proxy, &darray);
  CHECK(darray_str_size_get(&darray), 0);
  darray_str_resize(&darray, 8);
  CHECK(darray_str_size_get(&darray), 8);
  darray_str_resize(&darray, 0);
  CHECK(darray_str_size_get(&darray), 0);
  darray_str_resize(&darray, 33);
  CHECK(darray_str_size_get(&darray), 33);
  darray_str_release(&darray);

  check_memory_allocator(&allocator_proxy);
  mem_shutdown_proxy_allocator(&allocator_proxy);
}
#endif

#include "str.h"
#define DARRAY_NAME struct_str
#define DARRAY_DATA struct str
#define DARRAY_FUNCTOR_INIT str_init
#define DARRAY_FUNCTOR_COPY str_copy
#define DARRAY_FUNCTOR_RELEASE str_release
#define DARRAY_FUNCTOR_COPY_AND_RELEASE str_copy_and_release
#include "dynamic_array.h"

static void
test_struct(void)
{
  struct mem_allocator allocator_proxy;
  struct darray_struct_str darray, darray2;
  const size_t nstrs = sizeof(strs)/sizeof(const char*);
  size_t i = 0;

  mem_init_proxy_allocator(&allocator_proxy, &mem_default_allocator);

  darray_struct_str_init(&allocator_proxy, &darray);
  darray_struct_str_init(&allocator_proxy, &darray2);
  CHECK(darray_struct_str_size_get(&darray), 0);
  CHECK(darray_struct_str_size_get(&darray2), 0);

  darray_struct_str_resize(&darray, nstrs);
  FOR_EACH(i, 0, nstrs) {
    str_set(darray_struct_str_data_get(&darray) + i, strs[i]);
  }
  CHECK(darray_struct_str_size_get(&darray), nstrs);
  FOR_EACH(i, 0, nstrs) {
    const struct str* str = darray_struct_str_cdata_get(&darray) + i;
    CHECK(strcmp(str_cget(str), strs[i]), 0);
  }

  darray_struct_str_copy(&darray, &darray2);
  CHECK(darray_struct_str_size_get(&darray), 0);
  FOR_EACH(i, 0, nstrs) {
    struct str str;
    str_init(&allocator_proxy, &str);
    str_set(&str, strs[i]);
    darray_struct_str_push_back(&darray, &str);
    str_release(&str);
  }

  CHECK(darray_struct_str_size_get(&darray), nstrs);
  FOR_EACH(i, 0, nstrs) {
    const struct str* str = darray_struct_str_cdata_get(&darray) + i;
    CHECK(strcmp(str_cget(str), strs[i]), 0);
  }

  darray_struct_str_copy(&darray2, &darray);
  CHECK(darray_struct_str_size_get(&darray2), nstrs);
  FOR_EACH(i, 0, nstrs) {
    const struct str* str = darray_struct_str_cdata_get(&darray) + i;
    const struct str* str2 = darray_struct_str_cdata_get(&darray2) + i;
    CHECK(strcmp(str_cget(str), strs[i]), 0);
    CHECK(strcmp(str_cget(str2), strs[i]), 0);
  }
  darray_struct_str_clear(&darray2);
  CHECK(darray_struct_str_size_get(&darray2), 0);

  darray_struct_str_copy_and_clear(&darray2, &darray);
  CHECK(darray_struct_str_size_get(&darray2), nstrs);
  CHECK(darray_struct_str_size_get(&darray), 0);
  FOR_EACH(i, 0, nstrs) {
    const struct str* str = darray_struct_str_cdata_get(&darray2) + i;
    CHECK(strcmp(str_cget(str), strs[i]), 0);
  }

  darray_struct_str_copy_and_release(&darray, &darray2);
  FOR_EACH(i, 0, nstrs) {
    const struct str* str = darray_struct_str_cdata_get(&darray) + i;
    CHECK(strcmp(str_cget(str), strs[i]), 0);
  }

  darray_struct_str_release(&darray);

  check_memory_allocator(&allocator_proxy);
  mem_shutdown_proxy_allocator(&allocator_proxy);
}

int
main(int argc, char** argv)
{
  (void)argc, (void)argv;
  /*test_primitive_type();*/
  test_struct();
  CHECK(mem_allocated_size(), 0);
  return 0;
}

