/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "float2.h"
#include "rsys.h"
#include <float.h>

#define CHECK_F2(a, b)                                                         \
  {                                                                            \
    float* a__ = (a);                                                          \
    float* b__ = (b);                                                          \
    CHECK((a__)[0], (b__)[0]);                                                 \
    CHECK((a__)[1], (b__)[1]);                                                 \
  } (void) 0

int
main(int argc, char** argv)
{
  float a[2], b[2], dst[2], f, c[2];
  (void)argc, (void)argv;

  f2_set(a, f2_splat(c, -1.f));
  CHECK_F2(a, c);
  CHECK(a[0], -1.f);
  CHECK(a[1], -1.f);
  f2_set(a, f2(c, 0.f, 1.f));
  CHECK(a[0], 0.f);
  CHECK(a[1], 1.f);
  f2_splat(a, -2.f);
  CHECK(a[0], -2.f);
  CHECK(a[1], -2.f);

  f2_set(a, f2(c, -1.f, 2.f));
  CHECK_F2(f2_minus(b, a), f2(c, 1.f, -2.f));
  CHECK_F2(b, f2(c, 1.f, -2.f));

  CHECK_F2(f2_addf(dst, a, 1.f), f2(c, 0.f, 3.f));
  CHECK_F2(dst, f2(c, 0.f, 3.f));
  CHECK_F2(f2_add(dst, a, b), f2_splat(c, 0.f));
  CHECK_F2(dst, f2_splat(c, 0.f));
  CHECK_F2(f2_subf(dst, a, 1.f), f2(c, -2.f, 1.f));
  CHECK_F2(dst, f2(c, -2.f, 1.f));
  CHECK_F2(f2_sub(dst, a, b), f2(c, -2.f, 4.f));
  CHECK_F2(dst, f2(c, -2.f, 4.f));
  CHECK_F2(f2_mulf(dst, a, 2.f), f2(c, -2.f, 4.f));
  CHECK_F2(dst, f2(c, -2.f, 4.f));
  CHECK_F2(f2_mul(dst, a, b), f2(c, -1.f, -4.f));
  CHECK_F2(dst, f2(c, -1.f, -4.f));
  CHECK_F2(f2_div(dst, dst, a), f2(c, 1.f, -2.f));
  CHECK_F2(dst, f2(c, 1.f, -2.f));
  CHECK_F2(f2_divf(dst, a, 2.f), f2(c, -0.5f, 1.f));
  CHECK_F2(dst, f2(c, -0.5f, 1.f));

  f2_set(a, f2(c, 0.f, 1.f));
  f2_set(b, f2(c, 1.f, 2.f));
  CHECK_F2(f2_lerp(dst, a, b, 0.5f), f2(c, 0.5f, 1.5f));
  CHECK_F2(dst, f2(c, 0.5f, 1.5f));
  CHECK(f2_sum(b), 3.f);
  CHECK(f2_dot(a, b), 2.f);
  CHECK(eq_epsf(f2_len(a), (float)sqrt(1.0f), FLT_EPSILON), 1);

  CHECK(f2_is_normalized(b), 0);
  f = f2_normalize(dst, b);
  CHECK(f2_is_normalized(b), 0);
  NCHECK(f2_is_normalized(dst), 0);
  CHECK(eq_epsf(f, (float)sqrt(5.f), FLT_EPSILON), 1);
  CHECK(eq_epsf(dst[0], b[0] / f, FLT_EPSILON), 1);
  CHECK(eq_epsf(dst[1], b[1] / f, FLT_EPSILON), 1);

  CHECK(f2_eq(a, a), 1);
  CHECK(f2_eq(a, b), 0);
  CHECK(f2_eq(a, f2(c, b[0], a[1])), 0);
  CHECK(f2_eq(a, f2(c, a[0], b[1])), 0);

  f2_set(b, a);
  f2_add(b, b, f2_splat(c, FLT_EPSILON * 0.5f));
  CHECK(f2_eq(a, b), 0);
  CHECK(f2_eq_eps(a, b, FLT_EPSILON), 1);
  CHECK(f2_eq_eps(a, b, FLT_EPSILON * 0.25f), 0);

  f2_set(a, f2(c, 1.f, -2.f));
  f2_set(b, f2(c, 3.f, 1.f));
  CHECK(f2_cross(a, b), 7.f);

  f2_set(a, f2(c, 1.f, -2.f));
  f2_set(b, f2(c, -3.f, 1.f));
  CHECK_F2(f2_max(dst, a, b), f2(c, 1.f, 1.f));
  CHECK_F2(dst, f2(c, 1.f, 1.f));
  CHECK_F2(f2_min(dst, a, b), f2(c, -3.f, -2.f));
  CHECK_F2(dst, f2(c, -3.f, -2.f));
  return 0;
}

