/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "float22.h"
#include "rsys.h"
#include <float.h>

#define CHECK_F2(A, B)                                                         \
  {                                                                            \
    const float* a__ = (A);                                                    \
    const float* b__ = (B);                                                    \
    int i__;                                                                   \
    FOR_EACH(i__, 0, 2)                                                        \
      CHECK(a__[i__], b__[i__]);                                               \
  } (void)0
#define CHECK_F22(A, B)                                                        \
  {                                                                            \
    const float* a__ = (A);                                                    \
    const float* b__ = (B);                                                    \
    int i__;                                                                   \
    FOR_EACH(i__, 0, 4)                                                        \
      CHECK(a__[i__], b__[i__]);                                               \
  } (void)0

int
main(int argc, char** argv)
{
  float a[4], b[4], dst[4], c[4];
  int i;
  (void)argc, (void)argv;

  f22_set(c, f22_splat(a, -1.f));
  FOR_EACH(i, 0, 4) {
    CHECK(a[i], -1.f);
    CHECK(c[i], -1.f);
  }
  f22_set(a, f22(c, 0.f, 1.f, 2.f, 3.f));
  FOR_EACH(i, 0, 4) {
    CHECK(a[i], (float)i);
  }
  CHECK_F22(f22_set_identity(a), f22(c, 1.f, 0.f, 0.f, 1.f));
  CHECK_F22(a, f22(c, 1.f, 0.f, 0.f, 1.f));

  f22_splat(a, -1.f);
  CHECK_F22(f22_set_row(a, f2(c, 0.f, 1.f), 0), f22(c, 0.f, -1.f, 1.f, -1.f));
  CHECK_F22(a, f22(c, 0.f, -1.f, 1.f, -1.f));
  CHECK_F22(f22_set_row(a, f2(c, 2.f, 3.f), 1), f22(c, 0.f, 2.f, 1.f, 3.f));
  CHECK_F22(a, f22(c, 0.f, 2.f, 1.f, 3.f));

  CHECK_F22(f22_transpose(a, a), f22(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F22(a, f22(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F22(f22_transpose(b, a), f22(c, 0.f, 2.f, 1.f, 3.f));
  CHECK_F22(b, f22(c, 0.f, 2.f, 1.f, 3.f));

  f22_splat(a, -1.f);
  CHECK_F22(f22_set_col(a, f2(c, 0.f, 1.f), 0), f22(c, 0.f, 1.f, -1.f, -1.f));
  CHECK_F22(a, f22(c, 0.f, 1.f, -1.f, -1.f));
  CHECK_F22(f22_set_col(a, f2(c, 2.f, 3.f), 1), f22(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F22(a, f22(c, 0.f, 1.f, 2.f, 3.f));

  CHECK_F2(f22_row(b, a, 0), f2(c, 0.f, 2.f));
  CHECK_F2(f22_row(b, a, 1), f2(c, 1.f, 3.f));

  CHECK_F2(f22_col(b + 2, a, 0), f2(c, 0.f, 1.f));
  CHECK_F2(f22_col(b, a, 1), f2(c, 2.f, 3.f));

  CHECK_F2(f22_col_ptr(a, 0), f2(c, 0.f, 1.f));
  CHECK_F2(f22_col_ptr(a, 1), f2(c, 2.f, 3.f));
  CHECK_F2(f22_col_cptr(a, 0), f2(c, 0.f, 1.f));
  CHECK_F2(f22_col_cptr(a, 1), f2(c, 2.f, 3.f));

  f22(a, 1.f, 2.f, 3.f, 4.f);
  CHECK_F2(f22_mulf2(dst, a, f2(c, 1.f, 2.f)), f2(c, 7.f, 10.f));
  CHECK_F2(dst, f2(c, 7.f, 10.f));
  CHECK_F2(f2_mulf22(dst, f2(c, 1.f, 2.f), a), f2(c, 5.f, 11.f));
  CHECK_F2(dst, f2(c, 5.f, 11.f));
  CHECK_F22(f22_mul(dst, a, -1.f), f22(c, -1.f, -2.f, -3.f, -4.f));
  CHECK_F22(dst, f22(c, -1.f, -2.f, -3.f, -4.f));

  f22(b, 2.f, 9.f, 8.f, 1.f);
  CHECK_F22(f22_mulf22(dst, a, b), f22(c, 29.f, 40.f, 11.f, 20.0f));
  CHECK_F22(dst, f22(c, 29.f, 40.f, 11.f, 20.0f));
  CHECK_F22(f22_mulf22(b, a, b), f22(c, 29.f, 40.f, 11.f, 20.0f));
  CHECK_F22(b, f22(c, 29.f, 40.f, 11.f, 20.0f));

  f22(a, 0.f, 1.f, 2.f, 3.f);
  f22(b, 1.f, 2.f, 3.f, 4.f);
  CHECK_F22(f22_add(dst, a, b), f22(c, 1.f, 3.f, 5.f, 7.f));
  CHECK_F22(dst, f22(c, 1.f, 3.f, 5.f, 7.f));
  CHECK_F22(f22_sub(dst, dst, b), f22(c, 0.f, 1.f, 2.f, 3.f));
	CHECK_F22(dst, f22(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F22(f22_minus(a, b), f22(c,  -1.f, -2.f, -3.f, -4.f));
  CHECK_F22(a, f22(c,  -1.f, -2.f, -3.f, -4.f));

  f22_set(a, b);
  CHECK(f22_eq(a, b), 1);
  f22_add(a, a, f22_splat(c, FLT_EPSILON));
  CHECK(f22_eq(a, b), 0);
  CHECK(f22_eq_eps(a, b, FLT_EPSILON), 1);
  CHECK(f22_eq_eps(a, b, FLT_EPSILON * 0.9f), 0);

  f22_set(a, f22(c, 1, 3, 2, 4));
  CHECK(f22_det(a), -2.f);
  CHECK(f22_inverse(b, a), -2.f);
  CHECK_F22(b, f22(c, -2.f, 1.5f, 1.f, -0.5f));
  CHECK(f22_invtrans(a, a), -2.f);
  CHECK_F22(a, f22(c, -2.f, 1.f, 1.5f, -0.5f));

  return 0;
}

