/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "float4.h"
#include "rsys.h"
#include <float.h>

#define CHECK_F4(a, b)                                                         \
  {                                                                            \
    float* a__ = (a);                                                          \
    float* b__ = (b);                                                          \
    CHECK((a__)[0], (b__)[0]);                                                 \
    CHECK((a__)[1], (b__)[1]);                                                 \
    CHECK((a__)[2], (b__)[2]);                                                 \
    CHECK((a__)[3], (b__)[3]);                                                 \
  } (void) 0

int
main(int argc, char** argv)
{
  float a[4], b[4], dst[4], f, c[4];
  (void)argc, (void)argv;

  f4_set(a, f4_splat(c, -1.f));
  CHECK(a[0], -1.f);
  CHECK(a[1], -1.f);
  CHECK(a[2], -1.f);
  CHECK(a[3], -1.f);
  f4_set(a, f4(c, 0.f, 1.f, 2.f, 3.f));
  CHECK(a[0], 0.f);
  CHECK(a[1], 1.f);
  CHECK(a[2], 2.f);
  CHECK(a[3], 3.f);
  f4_splat(a, -2.f);
  CHECK(a[0], -2.f);
  CHECK(a[1], -2.f);
  CHECK(a[2], -2.f);
  CHECK(a[3], -2.f);

  f4_set(a, f4(c, -1.f, 2.f, -3.f, 4.f));
  CHECK_F4(f4_minus(b, a), f4(c, 1.f, -2.f, 3.f, -4.f));
  CHECK_F4(b, f4(c, 1.f, -2.f, 3.f, -4.f));

  CHECK_F4(f4_add(dst, a, b), f4_splat(c, 0.f));
  CHECK_F4(dst, f4_splat(c, 0.f));
  CHECK_F4(f4_addf(dst, a, 1.f), f4(c, 0.f, 3.f, -2.f, 5.f));
  CHECK_F4(dst, f4(c, 0.f, 3.F, -2.f, 5.f));
  CHECK_F4(f4_sub(dst, a, b), f4(c, -2.f, 4.f, -6.f, 8.f));
  CHECK_F4(dst, f4(c, -2.f, 4.f, -6.f, 8.f));
  CHECK_F4(f4_subf(dst, a, 1.f), f4(c, -2.f, 1.f, -4.f, 3.f));
  CHECK_F4(dst, f4(c, -2.f, 1.f, -4.f, 3.f));
  CHECK_F4(f4_mulf(dst, a, 2.f), f4(c, -2.f, 4.f, -6.f, 8.f));
  CHECK_F4(dst, f4(c, -2.f, 4.f, -6.f, 8.f));
  CHECK_F4(f4_mul(dst, a, b), f4(c, -1.f, -4.f, -9.f, -16.f));
  CHECK_F4(dst, f4(c, -1.f, -4.f, -9.f, -16.f));
  CHECK_F4(f4_div(dst, dst, a), f4(c, 1.f, -2.f, 3.f, -4.f));
  CHECK_F4(dst, f4(c, 1.f, -2.f, 3.f, -4.f));
  CHECK_F4(f4_divf(dst, a, 2.f), f4(c, -0.5f, 1.f, -1.5f, 2.f));
  CHECK_F4(dst, f4(c, -0.5f, 1.f, -1.5f, 2.f));

  f4_set(a, f4(c, 0.f, 1.f, 2.f, 4.f));
  f4_set(b, f4(c, 1.f, 2.f, -1.f, 1.f));
  CHECK_F4(f4_lerp(dst, a, b, 0.5f), f4(c, 0.5f, 1.5f, 0.5f, 2.5f));
  CHECK_F4(dst, f4(c, 0.5f, 1.5f, 0.5f, 2.5f));
  CHECK(f4_sum(b), 3.f);
  CHECK(f4_dot(a, b), 4.f);
  CHECK(eq_epsf(f4_len(a), (float)sqrt(21.0), FLT_EPSILON), 1);

  f4_set(a, f4(c, 0.f, 4.f, 2.f, 3.f));
  CHECK(f4_is_normalized(a), 0);
  f = f4_normalize(dst, a);
  CHECK(f4_is_normalized(a), 0);
  NCHECK(f4_is_normalized(dst), 0);
  CHECK(eq_epsf(f, (float)sqrt(29.0), FLT_EPSILON), 1);
  CHECK(eq_epsf(dst[0], a[0] / f, FLT_EPSILON), 1);
  CHECK(eq_epsf(dst[1], a[1] / f, FLT_EPSILON), 1);
  CHECK(eq_epsf(dst[2], a[2] / f, FLT_EPSILON), 1);
  CHECK(eq_epsf(dst[3], a[3] / f, FLT_EPSILON), 1);

  CHECK(f4_eq(a, a), 1);
  CHECK(f4_eq(a, b), 0);
  CHECK(f4_eq(a, f4(c, b[0], a[1], a[2], a[3])), 0);
  CHECK(f4_eq(a, f4(c, a[0], b[1], a[2], a[3])), 0);
  CHECK(f4_eq(a, f4(c, a[0], a[1], b[2], a[3])), 0);
  CHECK(f4_eq(a, f4(c, a[0], a[1], a[2], b[3])), 0);

  f4_set(b, a);
  f4_add(b, b, f4_splat(c, FLT_EPSILON * 0.5f));
  CHECK(f4_eq(a, b), 0);
  CHECK(f4_eq_eps(a, b, FLT_EPSILON), 1);
  CHECK(f4_eq_eps(a, b, FLT_EPSILON * 0.25f), 0);

  f4_set(a, f4(c, 1.f, 3.f, -2.f, 0.1f));
  f4_set(b, f4(c, -1.f, 3.1f, -2.1f, 1.f));
  CHECK_F4(f4_max(dst, a, b), f4(c, 1.f, 3.1f, -2.f, 1.f));
  CHECK_F4(dst, f4(c, 1.f, 3.1f, -2.f, 1.f));
  CHECK_F4(f4_min(dst, a, b), f4(c, -1.f, 3.f, -2.1f, 0.1f));
  CHECK_F4(dst, f4(c, -1.f, 3.f, -2.1f, 0.1f));
  return 0;
}
