/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s3d.h"
#include "test_s3d_utils.h"
#include "test_s3d_cbox.h"

#include <rsys/math.h>

static const float cube_verts[] = {
  5.f, 5.f, 5.f,
  6.f, 5.f, 5.f,
  5.f, 6.f, 5.f,
  6.f, 6.f, 5.f,
  5.f, 5.f, 6.f,
  6.f, 5.f, 6.f,
  5.f, 6.f, 6.f,
  6.f, 6.f, 6.f
};
static const unsigned cube_nverts = sizeof(cube_verts) / sizeof(float[3]);

/* Front faces are CW. The normals point into the cube */
static const unsigned cube_ids[] = {
  0, 2, 1, 1, 2, 3, /* Front */
  0, 4, 2, 2, 4, 6, /* Left */
  4, 5, 6, 6, 5, 7, /* Back */
  3, 7, 1, 1, 7, 5, /* Right */
  2, 6, 3, 3, 6, 7, /* Top */
  0, 1, 4, 4, 1, 5 /* Bottom */
};
static const unsigned cube_ntris = sizeof(cube_ids) / sizeof(unsigned[3]);

static void
cube_get_ids(const unsigned itri, unsigned ids[3], void* data)
{
  const unsigned id = itri * 3;
  CHECK(data, NULL);
  NCHECK(ids, NULL);
  CHECK(itri < cube_ntris, 1);
  ids[0] = cube_ids[id + 0];
  ids[1] = cube_ids[id + 1];
  ids[2] = cube_ids[id + 2];
}

static void
cube_get_pos(const unsigned ivert, float pos[3], void* data)
{
  const unsigned i = ivert*3;
  CHECK(data, NULL);
  NCHECK(pos, NULL);
  CHECK(ivert < cube_nverts, 1);
  pos[0] = cube_verts[i + 0];
  pos[1] = cube_verts[i + 1];
  pos[2] = cube_verts[i + 2];
}

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct s3d_primitive prims[10];
  struct s3d_device* dev, *dev2;
  struct s3d_scene* scn;
  struct s3d_scene* scn2;
  struct s3d_scene* scn3;
  struct s3d_vertex_data attribs;
  struct s3d_shape* shapes[4];
  const size_t nshapes = sizeof(shapes)/sizeof(struct s3d_shape*);
  void* data = (void*)&cbox_walls_desc;
  size_t i;
  size_t nprims;
  float area, volume, lower[3], upper[3];
  unsigned id;
  int mask;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHECK(s3d_device_create(NULL, &allocator, 0, &dev), RES_OK);
  FOR_EACH(i, 0, nshapes)
    CHECK(s3d_shape_create_mesh(dev, shapes + i), RES_OK);

  CHECK(s3d_scene_create(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_create(dev, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_create(NULL, &scn), RES_BAD_ARG);
  CHECK(s3d_scene_create(dev, &scn), RES_OK);
  CHECK(s3d_scene_create(dev, &scn2), RES_OK);
  CHECK(s3d_scene_create(dev, &scn3), RES_OK);

  CHECK(s3d_scene_attach_shape(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_attach_shape(scn, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_attach_shape(NULL, shapes[0]), RES_BAD_ARG);
  CHECK(s3d_scene_attach_shape(scn, shapes[0]), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, shapes[0]), RES_BAD_ARG);

  CHECK(s3d_scene_detach_shape(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_detach_shape(scn, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_detach_shape(NULL, shapes[0]), RES_BAD_ARG);
  CHECK(s3d_scene_detach_shape(scn, shapes[0]), RES_OK);
  CHECK(s3d_scene_detach_shape(scn, shapes[0]), RES_BAD_ARG);

  FOR_EACH(i, 1, nshapes) {
    CHECK(s3d_scene_attach_shape(scn, shapes[i]), RES_OK);
    CHECK(s3d_shape_ref_put(shapes[i]), RES_OK);
  }

  CHECK(s3d_scene_instantiate(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_instantiate(scn, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_instantiate(NULL, shapes + 1), RES_BAD_ARG);
  CHECK(s3d_scene_instantiate(scn, shapes + 1), RES_OK);

  CHECK(s3d_shape_get_id(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_shape_get_id(shapes[1], NULL), RES_BAD_ARG);
  CHECK(s3d_shape_get_id(NULL, &id), RES_BAD_ARG);
  CHECK(s3d_shape_get_id(shapes[1], &id), RES_OK);
  NCHECK(id, S3D_INVALID_ID);

  CHECK(s3d_scene_clear(NULL), RES_BAD_ARG);
  CHECK(s3d_scene_clear(scn), RES_OK);
  CHECK(s3d_scene_clear(scn), RES_OK);
  CHECK(s3d_scene_instantiate(scn, shapes + 2), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, shapes[2]), RES_BAD_ARG);

  CHECK(s3d_scene_get_session_mask(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_session_mask(scn, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_session_mask(NULL, &mask), RES_BAD_ARG);
  CHECK(s3d_scene_get_session_mask(scn, &mask), RES_OK);
  CHECK(mask, 0);

  CHECK(s3d_scene_attach_shape(scn, shapes[0]), RES_OK);
  CHECK(s3d_scene_begin_session(NULL, 0), RES_BAD_ARG);
  CHECK(s3d_scene_begin_session(scn, 0), RES_BAD_ARG);
  CHECK(s3d_scene_begin_session(scn, S3D_TRACE|S3D_GET_PRIMITIVE), RES_OK);
  CHECK(s3d_scene_begin_session(scn, S3D_TRACE), RES_BAD_OP);

  CHECK(s3d_scene_get_session_mask(scn, &mask), RES_OK);
  CHECK(mask & S3D_TRACE, S3D_TRACE);
  CHECK(mask & S3D_SAMPLE, 0);

  CHECK(s3d_scene_clear(scn), RES_BAD_OP);
  CHECK(s3d_scene_detach_shape(scn, shapes[0]), RES_BAD_OP);

  CHECK(s3d_scene_end_session(NULL), RES_BAD_ARG);
  CHECK(s3d_scene_end_session(scn), RES_OK);
  CHECK(s3d_scene_end_session(scn), RES_BAD_OP);

  CHECK(s3d_scene_get_session_mask(scn, &mask), RES_OK);
  CHECK(mask, 0);

  CHECK(s3d_scene_attach_shape(scn2, shapes[1]), RES_OK);
  CHECK(s3d_scene_attach_shape(scn2, shapes[2]), RES_OK);
  CHECK(s3d_scene_begin_session(scn2, S3D_SAMPLE|S3D_TRACE), RES_OK);

  CHECK(s3d_scene_compute_area(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_compute_area(scn2, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_compute_area(NULL, &area), RES_BAD_ARG);
  CHECK(s3d_scene_compute_area(scn2, &area), RES_OK);
  CHECK(area, 0.f);

  CHECK(s3d_scene_compute_volume(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_compute_volume(scn2, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_compute_volume(NULL, &volume), RES_BAD_ARG);
  CHECK(s3d_scene_compute_volume(scn2, &volume), RES_OK);
  CHECK(volume, 0.f);

  CHECK(s3d_scene_primitives_count(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_primitives_count(scn2, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_primitives_count(NULL, &nprims), RES_BAD_ARG);
  CHECK(s3d_scene_primitives_count(scn2, &nprims), RES_OK);
  CHECK(nprims, 0);

  CHECK(s3d_scene_get_aabb(NULL, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(scn2, NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(NULL, lower, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(scn2, lower, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(NULL, NULL, upper), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(scn2, NULL, upper), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(NULL, lower, upper), RES_BAD_ARG);
  CHECK(s3d_scene_get_aabb(scn2, lower, upper), RES_OK);
  CHECK(lower[0] > upper[0], 1);
  CHECK(lower[1] > upper[1], 1);
  CHECK(lower[2] > upper[2], 1);

  CHECK(s3d_scene_end_session(scn2), RES_OK);

  CHECK(s3d_scene_compute_area(scn2, &area), RES_BAD_OP);
  CHECK(s3d_scene_compute_volume(scn2, &volume), RES_BAD_OP);
  CHECK(s3d_scene_primitives_count(scn2, &nprims), RES_BAD_OP);
  CHECK(s3d_scene_get_aabb(scn2, lower, upper), RES_BAD_OP);

  CHECK(s3d_scene_instantiate(scn2, shapes + 3), RES_OK);
  CHECK(s3d_scene_attach_shape(scn3, shapes[3]), RES_OK);
  CHECK(s3d_scene_begin_session(scn3, S3D_SAMPLE|S3D_TRACE), RES_BAD_ARG);

  CHECK(s3d_scene_detach_shape(scn, shapes[0]), RES_OK);

  CHECK(s3d_shape_ref_put(shapes[0]), RES_OK);
  CHECK(s3d_shape_ref_put(shapes[1]), RES_OK);
  CHECK(s3d_shape_ref_put(shapes[2]), RES_OK);
  CHECK(s3d_shape_ref_put(shapes[3]), RES_OK);

  CHECK(s3d_scene_ref_get(NULL), RES_BAD_ARG);
  CHECK(s3d_scene_ref_get(scn), RES_OK);
  CHECK(s3d_scene_ref_put(NULL), RES_BAD_ARG);
  CHECK(s3d_scene_ref_put(scn), RES_OK);
  CHECK(s3d_scene_ref_put(scn), RES_OK);
  CHECK(s3d_scene_ref_put(scn2), RES_OK);
  CHECK(s3d_scene_ref_put(scn3), RES_OK);

  CHECK(s3d_scene_create(dev, &scn), RES_OK);
  CHECK(s3d_scene_create(dev, &scn2), RES_OK);

  attribs.type = S3D_FLOAT3;
  attribs.usage = S3D_POSITION;
  attribs.get = cbox_get_position;
  CHECK(s3d_shape_create_mesh(dev, shapes + 0), RES_OK);
  CHECK(s3d_mesh_setup_indexed_vertices(shapes[0], cbox_walls_ntris,
    cbox_get_ids, cbox_walls_nverts, &attribs, 1, data), RES_OK);
  CHECK(s3d_scene_attach_shape(scn, shapes[0]), RES_OK);

  CHECK(s3d_scene_begin_session(scn, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_compute_area(scn, &area), RES_OK);
  CHECK(eq_epsf(area, 1532296.f, 1.e-6f), 1);
  CHECK(s3d_scene_primitives_count(scn, &nprims), RES_OK);
  CHECK(nprims, 10);
  CHECK(s3d_scene_get_aabb(scn, lower, upper), RES_OK);
  CHECK(eq_epsf(lower[0], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(lower[1], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(lower[2], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[0], 552.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[1], 559.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[2], 548.f, 1.e-6f), 1);
  CHECK(s3d_scene_end_session(scn), RES_OK);

  CHECK(s3d_scene_instantiate(scn, shapes + 1), RES_OK);
  CHECK(s3d_scene_attach_shape(scn2, shapes[1]), RES_OK);

  CHECK(s3d_scene_begin_session(scn2, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_compute_area(scn2, &area), RES_OK);
  CHECK(eq_epsf(area, 1532296.f, 1.e-6f), 1);
  CHECK(s3d_scene_primitives_count(scn, &nprims), RES_OK);
  CHECK(nprims, 10);
  CHECK(s3d_scene_get_aabb(scn, lower, upper), RES_OK);
  CHECK(eq_epsf(lower[0], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(lower[1], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(lower[2], 0.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[0], 552.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[1], 559.f, 1.e-6f), 1);
  CHECK(eq_epsf(upper[2], 548.f, 1.e-6f), 1);
  CHECK(s3d_scene_end_session(scn2), RES_OK);

  CHECK(s3d_scene_begin_session(scn2, S3D_GET_PRIMITIVE), RES_OK);
  CHECK(s3d_scene_compute_area(scn2, &area), RES_OK);
  CHECK(eq_epsf(area, 1532296.f, 1.e-6f), 1);
  CHECK(s3d_scene_primitives_count(scn, &nprims), RES_OK);
  CHECK(nprims, 10);

  CHECK(s3d_scene_get_primitive(NULL, 11, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_primitive(scn2, 11, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_primitive(NULL, 0, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_primitive(scn2, 0, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_primitive(NULL, 11, prims + 0), RES_BAD_ARG);
  CHECK(s3d_scene_get_primitive(scn2, 11, prims + 0), RES_BAD_ARG);
  CHECK(s3d_scene_get_primitive(NULL, 0, prims + 0), RES_BAD_ARG);

  FOR_EACH(i, 0, nprims) {
    size_t j;
    CHECK(s3d_scene_get_primitive(scn2, (unsigned)i, prims + i), RES_OK);
    CHECK(S3D_PRIMITIVE_EQ(prims + i, &S3D_PRIMITIVE_NULL), 0);
    CHECK(prims[i].scene_prim_id, i);
    FOR_EACH(j, 0, i)
      CHECK(S3D_PRIMITIVE_EQ(prims + i, prims + j), 0);
  }
  CHECK(s3d_scene_end_session(scn2), RES_OK);

  attribs.type = S3D_FLOAT3;
  attribs.usage = S3D_POSITION;
  attribs.get = cube_get_pos;
  CHECK(s3d_mesh_setup_indexed_vertices
    (shapes[0], cube_ntris, cube_get_ids, cube_nverts, &attribs, 1, NULL), RES_OK);

  CHECK(s3d_scene_begin_session(scn, S3D_TRACE), RES_OK);
  CHECK(s3d_scene_compute_area(scn, &area), RES_OK);
  CHECK(eq_epsf(area, 6.f, 1.e-6f), 1);
  CHECK(s3d_scene_compute_volume(scn, &volume), RES_OK);
  CHECK(eq_epsf(volume, 1.f, 1.e-6f), 1);
  CHECK(s3d_scene_end_session(scn), RES_OK);

  CHECK(s3d_scene_get_device(NULL, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_device(scn, NULL), RES_BAD_ARG);
  CHECK(s3d_scene_get_device(NULL, &dev2), RES_BAD_ARG);
  CHECK(s3d_scene_get_device(scn, &dev2), RES_OK);
  CHECK(dev2, dev);
  CHECK(s3d_scene_get_device(scn2, &dev2), RES_OK);
  CHECK(dev2, dev);

  CHECK(s3d_scene_ref_put(scn), RES_OK);
  CHECK(s3d_scene_ref_put(scn2), RES_OK);
  CHECK(s3d_shape_ref_put(shapes[0]), RES_OK);
  CHECK(s3d_shape_ref_put(shapes[1]), RES_OK);

  CHECK(s3d_device_ref_put(dev), RES_OK);;

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);
  return 0;
}

